from autorag.nodes.passagefilter import recency_filter

from tests.autorag.nodes.passagefilter.test_passage_filter_base import contents_example, time_list, ids_example, \
    scores_example, base_passage_filter_test, base_passage_filter_node_test, project_dir_with_corpus, \
    previous_result


def test_recency_filter():
    original_recency_filter = recency_filter.__wrapped__
    contents_result, id_result, score_result = original_recency_filter \
        (contents_example, scores_example, ids_example, time_list, threshold="2021-06-30")
    assert id_result[0] == [ids_example[0][1], ids_example[0][2], ids_example[0][3]]
    assert id_result[1] == [ids_example[1][2], ids_example[1][3]]
    assert contents_result[0] == [contents_example[0][1], contents_example[0][2], contents_example[0][3]]
    assert contents_result[1] == [contents_example[1][2], contents_example[1][3]]
    assert score_result[0] == [0.8, 0.1, 0.5]
    assert score_result[1] == [0.7, 0.3]
    base_passage_filter_test(contents_result, id_result, score_result)


def test_recency_filter_all_filtered():
    original_recency_filter = recency_filter.__wrapped__
    contents_result, id_result, score_result = original_recency_filter \
        (contents_example, scores_example, ids_example, time_list, threshold="2040-06-30")
    assert id_result[0] == [ids_example[0][3]]
    assert id_result[1] == [ids_example[1][3]]
    assert contents_result[0] == [contents_example[0][3]]
    assert contents_result[1] == [contents_example[1][3]]
    assert score_result[0] == [0.5]
    assert score_result[1] == [0.3]
    base_passage_filter_test(contents_result, id_result, score_result)


def test_recency_filter_wrong_threshold():
    original_recency_filter = recency_filter.__wrapped__
    contents_result, id_result, score_result = original_recency_filter \
        (contents_example, scores_example, ids_example, time_list, threshold="havertz")
    assert id_result[0] == ids_example[0]
    assert id_result[1] == ids_example[1]
    assert contents_result[0] == contents_example[0]
    assert contents_result[1] == contents_example[1]
    assert score_result[0] == scores_example[0]
    assert score_result[1] == scores_example[1]
    base_passage_filter_test(contents_result, id_result, score_result)


def test_recency_filter_minutes():
    original_recency_filter = recency_filter.__wrapped__
    contents_result, id_result, score_result = original_recency_filter \
        (contents_example, scores_example, ids_example, time_list, threshold="2022-03-05 12:35")
    assert id_result[0] == [ids_example[0][3]]
    assert id_result[1] == [ids_example[1][2], ids_example[1][3]]
    assert contents_result[0] == [contents_example[0][3]]
    assert contents_result[1] == [contents_example[1][2], contents_example[1][3]]
    assert score_result[0] == [0.5]
    assert score_result[1] == [0.7, 0.3]
    base_passage_filter_test(contents_result, id_result, score_result)


def test_recency_filter_seconds():
    original_recency_filter = recency_filter.__wrapped__
    contents_result, id_result, score_result = original_recency_filter \
        (contents_example, scores_example, ids_example, time_list, threshold="2022-03-05 12:40:30")
    assert id_result[0] == [ids_example[0][3]]
    assert id_result[1] == [ids_example[1][3]]
    assert contents_result[0] == [contents_example[0][3]]
    assert contents_result[1] == [contents_example[1][3]]
    assert score_result[0] == [0.5]
    assert score_result[1] == [0.3]
    base_passage_filter_test(contents_result, id_result, score_result)


def test_recency_filter_node(project_dir_with_corpus):
    result_df = recency_filter(
        project_dir=project_dir_with_corpus, previous_result=previous_result, threshold="2021-06-30")
    base_passage_filter_node_test(result_df)



import pytest

from autorag.nodes.passagereranker import tart
from tests.autorag.nodes.passagereranker.test_passage_reranker_base import base_reranker_test, queries_example, \
    contents_example, scores_example, ids_example, base_reranker_node_test, project_dir, previous_result
from tests.delete_tests import is_github_action


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_tart():
    top_k = 1
    original_tart = tart.__wrapped__
    contents_result, id_result, score_result \
        = original_tart(queries_example, contents_example, scores_example, ids_example, top_k)
    base_reranker_test(contents_result, id_result, score_result, top_k)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_tart_node():
    top_k = 1
    result_df = tart(project_dir=project_dir, previous_result=previous_result, top_k=top_k)
    base_reranker_node_test(result_df, top_k)



from typing import List

from autorag.nodes.passagefilter.base import passage_filter_node


@passage_filter_node
def pass_passage_filter(queries: List[str], contents_list: List[List[str]],
                        scores_list: List[List[float]], ids_list: List[List[str]]):
    """
    Do not perform filtering.
    Return given passages, scores, and ids as is.
    """
    return contents_list, ids_list, scores_list



import functools
from typing import List, Callable, Any, Tuple

import pandas as pd

from autorag.evaluate.metric import retrieval_token_f1, retrieval_token_precision, retrieval_token_recall


def evaluate_retrieval_contents(retrieval_gt: List[List[str]], metrics: List[str]):
    def decorator_evaluate_retireval_contents(
            func: Callable[[Any], Tuple[List[List[str]], List[List[str]], List[List[float]]]]):
        """
        Decorator for evaluating retrieval contents.
        You can use this decorator to any method that returns (contents, scores, ids),
        which is the output of conventional retrieval modules.

        :param func: Must return (contents, scores, ids)
        :return: pd.DataFrame, which is the evaluation result and function result.
        """

        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> pd.DataFrame:
            contents, pred_ids, scores = func(*args, **kwargs)
            metric_funcs = {
                retrieval_token_recall.__name__: retrieval_token_recall,
                retrieval_token_precision.__name__: retrieval_token_precision,
                retrieval_token_f1.__name__: retrieval_token_f1,
            }

            metrics_scores = {}
            for metric in metrics:
                if metric not in metric_funcs:
                    raise ValueError(f"metric {metric} is not in supported metrics: {metric_funcs.keys()}")
                else:
                    metric_func = metric_funcs[metric]
                    metric_scores = metric_func(gt_contents=retrieval_gt, pred_contents=contents)
                    metrics_scores[metric] = metric_scores

            metric_result_df = pd.DataFrame(metrics_scores)
            execution_result_df = pd.DataFrame({
                'retrieved_contents': contents,
                'retrieved_ids': pred_ids,
                'retrieve_scores': scores,
            })
            result_df = pd.concat([execution_result_df, metric_result_df], axis=1)
            return result_df

        return wrapper

    return decorator_evaluate_retireval_contents



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from typing import Any, Dict, List, Optional

from transformers import T5Tokenizer


class EncT5Tokenizer(T5Tokenizer):
    def __init__(
            self,
            vocab_file,
            bos_token="<s>",
            eos_token="</s>",
            unk_token="<unk>",
            pad_token="<pad>",
            extra_ids=100,
            additional_special_tokens=None,
            sp_model_kwargs: Optional[Dict[str, Any]] = None,
            **kwargs,
    ) -> None:
        sp_model_kwargs = {} if sp_model_kwargs is None else sp_model_kwargs

        super().__init__(
            vocab_file=vocab_file,
            bos_token=bos_token,
            eos_token=eos_token,
            unk_token=unk_token,
            pad_token=pad_token,
            extra_ids=extra_ids,
            additional_special_tokens=additional_special_tokens,
            sp_model_kwargs=sp_model_kwargs,
            **kwargs,
        )

    def get_special_tokens_mask(
            self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None,
            already_has_special_tokens: bool = False
    ) -> List[int]:
        """
        Retrieve sequence ids from a token list that has no special tokens added. This method is called when adding
        special tokens using the tokenizer `prepare_for_model` method.
        Args:
            token_ids_0 (`List[int]`):
                List of IDs.
            token_ids_1 (`List[int]`, *optional*):
                Optional second list of IDs for sequence pairs.
            already_has_special_tokens (`bool`, *optional*, defaults to `False`):
                Whether or not the token list is already formatted with special tokens for the model.
        Returns:
            `List[int]`: A list of integers in the range [0, 1]: 1 for a special token, 0 for a sequence token.
        """
        if already_has_special_tokens:
            return super().get_special_tokens_mask(
                token_ids_0=token_ids_0, token_ids_1=token_ids_1, already_has_special_tokens=True
            )

        # normal case: some special tokens
        if token_ids_1 is None:
            return [1] + ([0] * len(token_ids_0)) + [1]
        return [1] + ([0] * len(token_ids_0)) + [1] + ([0] * len(token_ids_1)) + [1]

    def create_token_type_ids_from_sequences(
            self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        """
        Create a mask from the two sequences passed to be used in a sequence-pair classification task. T5 does not make
        use of token type ids, therefore a list of zeros is returned.
        Args:
            token_ids_0 (`List[int]`):
                List of IDs.
            token_ids_1 (`List[int]`, *optional*):
                Optional second list of IDs for sequence pairs.
        Returns:
            `List[int]`: List of zeros.
        """
        bos = [self.bos_token_id]
        eos = [self.eos_token_id]

        if token_ids_1 is None:
            return len(bos + token_ids_0 + eos) * [0]
        return len(bos + token_ids_0 + eos + token_ids_1 + eos) * [0]

    def build_inputs_with_special_tokens(
            self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        """
        Build model inputs from a sequence or a pair of sequence for sequence classification tasks by concatenating and
        adding special tokens. A sequence has the following format:
        - single sequence: `<s> X </s>`
        - pair of sequences: `<s> A </s> B </s>`
        Args:
            token_ids_0 (`List[int]`):
                List of IDs to which the special tokens will be added.
            token_ids_1 (`List[int]`, *optional*):
                Optional second list of IDs for sequence pairs.
        Returns:
            `List[int]`: List of [input IDs](../glossary#input-ids) with the appropriate special tokens.
        """
        if token_ids_1 is None:
            return [self.bos_token_id] + token_ids_0 + [self.eos_token_id]
        else:
            return [self.bos_token_id] + token_ids_0 + [self.eos_token_id] + token_ids_1 + [self.eos_token_id]



import pytest

from autorag.nodes.passagereranker import monot5
from tests.autorag.nodes.passagereranker.test_passage_reranker_base \
    import (base_reranker_test, base_reranker_node_test, queries_example,
            contents_example, scores_example, ids_example, project_dir, previous_result)
from tests.delete_tests import is_github_action


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_monot5():
    top_k = 1
    original_monot5 = monot5.__wrapped__
    contents_result, id_result, score_result \
        = original_monot5(queries_example, contents_example, scores_example, ids_example, top_k)
    base_reranker_test(contents_result, id_result, score_result, top_k)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_monot5_node():
    top_k = 1
    result_df = monot5(project_dir=project_dir, previous_result=previous_result, top_k=top_k)
    base_reranker_node_test(result_df, top_k)



from autorag.support import get_support_modules, get_support_nodes


def test_get_support_modules():
    result = get_support_modules('bm25')
    assert result.__name__ == 'bm25'


def test_get_support_nodes():
    result = get_support_nodes('retrieval')
    assert result.__name__ == 'run_retrieval_node'



from typing import List, Tuple

import pandas as pd
import torch
from FlagEmbedding import FlagLLMReranker

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.nodes.passagereranker.flag_embedding import flag_embedding_run_model
from autorag.utils.util import flatten_apply, sort_by_scores, select_top_k


@passage_reranker_node
def flag_embedding_llm_reranker(queries: List[str], contents_list: List[List[str]],
                                scores_list: List[List[float]], ids_list: List[List[str]],
                                top_k: int, batch: int = 64, use_fp16: bool = False,
                                model_name: str = "BAAI/bge-reranker-v2-gemma",
                                ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents based on their relevance to a query using BAAI LLM-based-Reranker model.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param batch: The number of queries to be processed in a batch
    :param use_fp16: Whether to use fp16 for inference
    :param model_name: The name of the BAAI Reranker LLM-based-model name.
        Default is "BAAI/bge-reranker-v2-gemma"
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    model = FlagLLMReranker(
        model_name_or_path=model_name, use_fp16=use_fp16
    )
    nested_list = [list(map(lambda x: [query, x], content_list)) for query, content_list in zip(queries, contents_list)]
    rerank_scores = flatten_apply(flag_embedding_run_model, nested_list, model=model, batch_size=batch)

    df = pd.DataFrame({
        'contents': contents_list,
        'ids': ids_list,
        'scores': rerank_scores,
    })
    df[['contents', 'ids', 'scores']] = df.apply(sort_by_scores, axis=1, result_type='expand')
    results = select_top_k(df, ['contents', 'ids', 'scores'], top_k)

    del model
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return results['contents'].tolist(), results['ids'].tolist(), results['scores'].tolist()



import os
import pathlib
import re
from unittest.mock import patch

import pandas as pd
import pytest
from llama_index.core.base.llms.types import CompletionResponse
from llama_index.core.llms import MockLLM

from autorag.data.qacreation.llama_index import async_qa_gen_llama_index, generate_qa_llama_index_by_ratio, \
    generate_qa_llama_index

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent
resource_dir = os.path.join(root_dir, "resources")

content = """Fire and brimstone ( or , alternatively , brimstone and fire ) is an idiomatic expression of referring to God 's wrath in the Hebrew Bible ( Old Testament ) and the New Testament .
In the Bible , it often appears in reference to the fate of the unfaithful .
Brimstone , an archaic term synonymous with sulfur , evokes the acrid odor of sulphur dioxide given off by lightning strikes .
Lightning was understood as divine punishment by many ancient religions ; the association of sulphur with God 's retribution is common in the Bible .
The English phrase `` fire and brimstone '' originates in the King James Bible ."""

prompt_dir = os.path.join(resource_dir, "qa_gen_prompts")
sample_prompt = open(os.path.join(prompt_dir, "prompt1.txt")).read()


@pytest.fixture
def contents():
    df = pd.read_csv(os.path.join(resource_dir, "sample_contents_nqa.csv"))
    yield df['passage'].tolist()[:6]


async def acomplete_qa_creation(self, messages, **kwargs):
    pattern = r'Number of questions to generate: (\d+)'
    matches = re.findall(pattern, messages)
    num_questions = int(matches[-1])
    return CompletionResponse(text=
                              "[Q]: Is this the test question?\n[A]: Yes, this is the test answer." * num_questions)


@patch.object(MockLLM, "acomplete", acomplete_qa_creation)
@pytest.mark.asyncio()
async def test_async_qa_gen_llama_index():
    result = await async_qa_gen_llama_index(content, llm=MockLLM(), prompt=sample_prompt, question_num=3)
    assert len(result) == 3
    for res in result:
        assert "query" in res
        assert "generation_gt" in res
        assert bool(res['query'])
        assert bool(res['generation_gt'])


@patch.object(MockLLM, "acomplete", acomplete_qa_creation)
def test_qa_gen_llama_index(contents):
    llm = MockLLM()
    result = generate_qa_llama_index(llm, contents, sample_prompt)
    check_multi_qa_gen(result)


@patch.object(MockLLM, "acomplete", acomplete_qa_creation)
def test_qa_gen_llama_index_by_ratio(contents):
    ratio_dict = {
        str(os.path.join(prompt_dir, "prompt1.txt")): 1,
        str(os.path.join(prompt_dir, "prompt2.txt")): 2,
        str(os.path.join(prompt_dir, "prompt3.txt")): 3,
    }
    llm = MockLLM()
    result = generate_qa_llama_index_by_ratio(llm, contents, ratio_dict, batch=8)
    check_multi_qa_gen(result)


def check_multi_qa_gen(result):
    assert len(result) == 6
    for res in result:
        assert all("query" in r for r in res)
        assert all("generation_gt" in r for r in res)
        assert all(bool(r['query']) for r in res)
        assert all(bool(r['generation_gt']) for r in res)



from copy import deepcopy
from typing import Union, List, Dict, Tuple, Any

from autorag import embedding_models


def cast_metrics(metrics: Union[List[str], List[Dict]]) -> Tuple[List[str], List[Dict[str, Any]]]:
    """
     Turn metrics to list of metric names and parameter list.

    :param metrics: List of string or dictionary.
    :return: The list of metric names and dictionary list of metric parameters.
    """
    metrics_copy = deepcopy(metrics)
    if not isinstance(metrics_copy, list):
        raise ValueError("metrics must be a list of string or dictionary.")
    if isinstance(metrics_copy[0], str):
        return metrics_copy, [{} for _ in metrics_copy]
    elif isinstance(metrics_copy[0], dict):
        # pop 'metric_name' key from dictionary
        metric_names = list(map(lambda x: x.pop('metric_name'), metrics_copy))
        metric_params = [dict(map(lambda x, y: cast_embedding_model(x, y), metric.keys(), metric.values())) for metric
                         in metrics_copy]
        return metric_names, metric_params
    else:
        raise ValueError("metrics must be a list of string or dictionary.")


def cast_embedding_model(key, value):
    if key == 'embedding_model':
        return key, embedding_models[value]
    else:
        return key, value



from unittest.mock import patch

import pandas as pd
import pytest
from openai.resources.chat import Completions
from openai.types.chat import ChatCompletion, ChatCompletionMessage, ChatCompletionTokenLogprob
from openai.types.chat.chat_completion import Choice, ChoiceLogprobs
from openai.types.chat.chat_completion_token_logprob import TopLogprob
from transformers import AutoTokenizer

from autorag.evaluate.generation import evaluate_generation

generation_gts = [
    ['The dog had bit the man.', 'The man had bitten the dog.'],
    ['I want to be a artist, but I end up to be a programmer.'],
    ['To be a artist these days, you can overcome by AI.',
     'To be a programmer these days, you can overcome by AI.',
     'To be a lawyer these days, you can overcome by AI.'],
]
pseudo_generations = [
    'The dog bit the man.',
    'It really like to be a programmer, but I think artist is my passion.',
    'To be a artist these days, you can overcome by AI.',
]

tokenizer = AutoTokenizer.from_pretrained('gpt2')
pseudo_tokens = list(map(lambda x: tokenizer.tokenize(x), pseudo_generations))
pseudo_log_probs = list(map(lambda x: [0.1] * len(x), pseudo_tokens))


@evaluate_generation(generation_gt=generation_gts, metrics=['bleu', 'meteor', 'rouge', 'sem_score', 'g_eval'])
def pseudo_generation():
    return pseudo_generations


@evaluate_generation(generation_gt=generation_gts, metrics=['bleu', 'meteor', 'donggeon_metric'])
def pseudo_generation_with_log_probs():
    return pseudo_generations, pseudo_tokens, pseudo_log_probs


@evaluate_generation(generation_gt=generation_gts, metrics=[{'metric_name': 'bleu'}, {'metric_name': 'sem_score',
                                                                                      'embedding_model': 'openai'}])
def pseudo_generation_dict_metrics():
    return pseudo_generations


def mock_g_eval_openai_create(*args, **kwargs):
    sample_choice = Choice(finish_reason="stop", index=0,
                           message=ChatCompletionMessage(
                               content="2",
                               role="assistant",
                           ),
                           logprobs=ChoiceLogprobs(
                               content=[
                                   ChatCompletionTokenLogprob(token="2",
                                                              logprob=2.8,
                                                              top_logprobs=[
                                                                  TopLogprob(token="2", logprob=2.8),
                                                              ])
                               ],
                           ))
    if 'n' not in kwargs.keys():
        n = 1
    else:
        n = kwargs['n']
    return ChatCompletion(id='_id',
                          choices=[sample_choice] * n,
                          created=1713363661, model=kwargs['model'],
                          object="chat.completion")


@patch.object(Completions, 'create', mock_g_eval_openai_create)
def test_evaluate_generation():
    result_df = pseudo_generation()
    assert isinstance(result_df, pd.DataFrame)
    assert len(result_df) == 3
    assert len(result_df.columns) == 6
    assert set(result_df.columns) == {'generated_texts', 'bleu', 'meteor', 'rouge', 'sem_score', 'g_eval'}

    with pytest.warns():
        result_df_log_probs = pseudo_generation_with_log_probs()
    assert isinstance(result_df_log_probs, pd.DataFrame)
    assert len(result_df_log_probs) == 3
    assert len(result_df_log_probs.columns) == 5
    assert set(result_df_log_probs.columns) == {'generated_texts', 'bleu', 'meteor', 'generated_tokens',
                                                'generated_log_probs'}

    assert result_df_log_probs['generated_texts'].tolist() == pseudo_generations
    assert result_df_log_probs['generated_tokens'].tolist() == pseudo_tokens
    assert result_df_log_probs['generated_log_probs'].tolist() == pseudo_log_probs

    assert all(list(map(lambda x: x[0] == pytest.approx(x[1], 0.001),
                        zip(result_df['bleu'].tolist(), [51.1507, 23.5783, 100.0]))))
    assert all(list(map(lambda x: x[0] == pytest.approx(x[1], 0.001),
                        zip(result_df['meteor'].tolist(), [0.853462, 0.5859375, 1.0]))))
    assert all(list(map(lambda x: x[0] == pytest.approx(x[1], 0.001),
                        zip(result_df['rouge'].tolist(), [0.909, 0.35714, 1.0]))))

    result_df = pseudo_generation_dict_metrics()
    assert isinstance(result_df, pd.DataFrame)
    assert len(result_df) == 3
    assert len(result_df.columns) == 3
    assert set(result_df.columns) == {'generated_texts', 'bleu', 'sem_score'}



import os
import pathlib
import tempfile

import pandas as pd
import pytest
import yaml
from fastapi.testclient import TestClient

from autorag.deploy import summary_df_to_yaml, extract_best_config, Runner, extract_node_line_names, \
    extract_node_strategy
from autorag.evaluator import Evaluator
from tests.delete_tests import is_github_action

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent
resource_dir = os.path.join(root_dir, 'resources')


@pytest.fixture
def evaluator():
    with tempfile.TemporaryDirectory() as project_dir:
        evaluator = Evaluator(os.path.join(resource_dir, 'qa_data_sample.parquet'),
                              os.path.join(resource_dir, 'corpus_data_sample.parquet'),
                              project_dir=project_dir)
        yield evaluator


@pytest.fixture
def full_config():
    yaml_path = os.path.join(resource_dir, 'full.yaml')
    with open(yaml_path, 'r') as f:
        yaml_dict = yaml.safe_load(f)
    return yaml_dict


summary_df = pd.DataFrame({
    'node_line_name': ['node_line_2', 'node_line_2', 'node_line_1'],
    'node_type': ['retrieval', 'rerank', 'generation'],
    'best_module_filename': ['bm25=>top_k_50.parquet', 'upr=>model_llama-2-havertz_chelsea.parquet',
                             'gpt-4=>top_p_0.9.parquet'],
    'best_module_name': ['bm25', 'upr', 'gpt-4'],
    'best_module_params': [{'top_k': 50}, {'model': 'llama-2', 'havertz': 'chelsea'}, {'top_p': 0.9}],
    'best_execution_time': [1.0, 0.5, 2.0]
})
solution_dict = {
    'node_lines': [
        {
            'node_line_name': 'node_line_2',
            'nodes': [
                {
                    'node_type': 'retrieval',
                    'strategy': {
                        'metrics': ['retrieval_f1', 'retrieval_recall', 'retrieval_precision'],
                    },
                    'modules': [
                        {
                            'module_type': 'bm25',
                            'top_k': 50
                        }
                    ]
                },
                {
                    'node_type': 'rerank',
                    'strategy': {
                        'metrics': ['retrieval_f1', 'retrieval_recall', 'retrieval_precision'],
                        'speed_threshold': 10,
                    },
                    'modules': [
                        {
                            'module_type': 'upr',
                            'model': 'llama-2',
                            'havertz': 'chelsea'
                        }
                    ]
                }
            ]
        },
        {
            'node_line_name': 'node_line_1',
            'nodes': [
                {
                    'node_type': 'generation',
                    'strategy': {
                        'metrics': ['bleu', 'rouge'],
                    },
                    'modules': [
                        {
                            'module_type': 'gpt-4',
                            'top_p': 0.9
                        }
                    ]
                }
            ]
        }
    ]
}


@pytest.fixture
def pseudo_trial_path():
    with tempfile.TemporaryDirectory() as project_dir:
        trial_path = os.path.join(project_dir, '0')
        os.makedirs(trial_path)
        summary_df.to_csv(os.path.join(trial_path, 'summary.csv'), index=False)
        with open(os.path.join(trial_path, 'config.yaml'), 'w') as f:
            yaml.dump(solution_dict, f)
        yield trial_path


def test_extract_node_line_names(full_config):
    node_line_names = extract_node_line_names(full_config)
    assert node_line_names == ['pre_retrieve_node_line', 'retrieve_node_line', 'post_retrieve_node_line']


def test_extract_node_strategy(full_config):
    node_strategies = extract_node_strategy(full_config)
    assert set(list(node_strategies.keys())) == {
        'query_expansion', 'retrieval', 'passage_reranker', 'passage_compressor',
        'prompt_maker', 'generator'
    }
    assert node_strategies['retrieval'] == {
        'metrics': ['retrieval_f1', 'retrieval_recall', 'retrieval_precision'],
        'speed_threshold': 10,
    }


def test_summary_df_to_yaml():
    yaml_dict = summary_df_to_yaml(summary_df, solution_dict)
    assert yaml_dict == solution_dict


def test_extract_best_config(pseudo_trial_path):
    yaml_dict = extract_best_config(pseudo_trial_path)
    assert yaml_dict == solution_dict
    with tempfile.NamedTemporaryFile(suffix='yaml', mode='w+t') as yaml_path:
        yaml_dict = extract_best_config(pseudo_trial_path, yaml_path.name)
        assert yaml_dict == solution_dict
        assert os.path.exists(yaml_path.name)
        yaml_dict = yaml.safe_load(yaml_path)
        assert yaml_dict == solution_dict


def test_runner(evaluator):
    os.environ['BM25'] = 'bm25'
    evaluator.start_trial(os.path.join(resource_dir, 'simple.yaml'))
    project_dir = evaluator.project_dir

    def runner_test(runner: Runner):
        answer = runner.run('What is the best movie in Korea? Have Korea movie ever won Oscar?',
                            'retrieved_contents')
        assert len(answer) == 1
        assert isinstance(answer, list)
        assert isinstance(answer[0], str)

    runner = Runner.from_trial_folder(os.path.join(project_dir, '0'))
    runner_test(runner)
    runner_test(runner)

    with tempfile.NamedTemporaryFile(suffix='yaml', mode='w+t') as yaml_path:
        extract_best_config(os.path.join(project_dir, '0'), yaml_path.name)
        runner = Runner.from_yaml(yaml_path.name, project_dir=project_dir)
        runner_test(runner)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_runner_full(evaluator):
    runner = Runner.from_trial_folder(os.path.join(resource_dir, 'result_project', '0'))
    answer = runner.run('What is the best movie in Korea? Have Korea movie ever won Oscar?')
    assert isinstance(answer, str)
    assert bool(answer)


def test_runner_api_server(evaluator):
    os.environ['BM25'] = 'bm25'
    project_dir = evaluator.project_dir
    import nest_asyncio
    nest_asyncio.apply()
    evaluator.start_trial(os.path.join(resource_dir, 'simple.yaml'))
    runner = Runner.from_trial_folder(os.path.join(project_dir, '0'))

    client = TestClient(runner.app)

    # Use the TestClient to make a request to the server
    response = client.post('/run', json={
        'query': 'What is the best movie in Korea? Have Korea movie ever won Oscar?',
        'result_column': 'retrieved_contents'
    })
    assert response.status_code == 200
    assert 'retrieved_contents' in response.json()
    retrieved_contents = response.json()['retrieved_contents']
    assert len(retrieved_contents) == 1
    assert isinstance(retrieved_contents, list)
    assert isinstance(retrieved_contents[0], str)



import asyncio
from typing import List

from llama_index.core.service_context_elements.llm_predictor import LLMPredictorType

from autorag.nodes.queryexpansion.base import query_expansion_node
from autorag.utils.util import process_batch

decompose_prompt = """Decompose a question in self-contained sub-questions. Use \"The question needs no decomposition\" when no decomposition is needed.

    Example 1:

    Question: Is Hamlet more common on IMDB than Comedy of Errors?
    Decompositions: 
    1: How many listings of Hamlet are there on IMDB?
    2: How many listing of Comedy of Errors is there on IMDB?

    Example 2:

    Question: Are birds important to badminton?

    Decompositions:
    The question needs no decomposition

    Example 3:

    Question: Is it legal for a licensed child driving Mercedes-Benz to be employed in US?

    Decompositions:
    1: What is the minimum driving age in the US?
    2: What is the minimum age for someone to be employed in the US?

    Example 4:

    Question: Are all cucumbers the same texture?

    Decompositions:
    The question needs no decomposition

    Example 5:

    Question: Hydrogen's atomic number squared exceeds number of Spice Girls?

    Decompositions:
    1: What is the atomic number of hydrogen?
    2: How many Spice Girls are there?

    Example 6:

    Question: {question}

    Decompositions:"
    """


@query_expansion_node
def query_decompose(queries: List[str], llm: LLMPredictorType,
                    prompt: str = decompose_prompt,
                    batch: int = 16) -> List[List[str]]:
    """
    decompose query to little piece of questions.
    :param queries: List[str], queries to decompose.
    :param llm: LLMPredictorType, language model to use.
    :param prompt: str, prompt to use for query decomposition.
        default prompt comes from Visconde's StrategyQA few-shot prompt.
    :param batch: int, batch size for llm.
        Default is 16.
    :return: List[List[str]], list of decomposed query. Return input query if query is not decomposable.
    """
    # Run async query_decompose_pure function
    tasks = [query_decompose_pure(query, llm, prompt) for query in queries]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(process_batch(tasks, batch_size=batch))
    return results


async def query_decompose_pure(query: str, llm: LLMPredictorType,
                               prompt: str = decompose_prompt) -> List[str]:
    """
    decompose query to little piece of questions.
    :param query: str, query to decompose.
    :param llm: LLMPredictorType, language model to use.
    :param prompt: str, prompt to use for query decomposition.
        default prompt comes from Visconde's StrategyQA few-shot prompt.
    :return: List[str], list of a decomposed query. Return input query if query is not decomposable.
    """
    if prompt == "":
        prompt = decompose_prompt
    full_prompt = "prompt: " + prompt + "\n\n" "question: " + query
    answer = await llm.acomplete(full_prompt)
    if answer.text == "the question needs no decomposition.":
        return [query]
    try:
        lines = [line.strip() for line in answer.text.splitlines() if line.strip()]
        if lines[0].startswith("Decompositions:"):
            lines.pop(0)
        questions = [line.split(':', 1)[1].strip() for line in lines if ':' in line]
        if not questions:
            return [query]
        return questions
    except:
        return [query]



from typing import List, Tuple

import pandas as pd
import torch
from transformers import AutoModel, AutoTokenizer

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.utils.util import make_batch, select_top_k, flatten_apply, sort_by_scores


@passage_reranker_node
def colbert_reranker(queries: List[str], contents_list: List[List[str]],
                     scores_list: List[List[float]], ids_list: List[List[str]],
                     top_k: int, batch: int = 64,
                     model_name: str = "colbert-ir/colbertv2.0",
                     ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents with Colbert rerank models.
    You can get more information about a Colbert model at https://huggingface.co/colbert-ir/colbertv2.0.
    It uses BERT-based model, so recommend using CUDA gpu for faster reranking.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param batch: The number of queries to be processed in a batch
        Default is 64.
    :param model_name: The model name for Colbert rerank.
        You can choose colbert model for reranking.
        Default is "colbert-ir/colbertv2.0".
    :return: Tuple of lists containing the reranked contents, ids, and scores
    """
    device = "cuda" if torch.cuda.is_available() else "cpu"
    model = AutoModel.from_pretrained(model_name).to(device)
    tokenizer = AutoTokenizer.from_pretrained(model_name)

    nested_list = [list(map(lambda x: [query, x], content_list)) for query, content_list in zip(queries, contents_list)]

    rerank_scores = flatten_apply(colbert_run_model, nested_list, model=model,
                                  tokenizer=tokenizer, batch_size=batch, device=device)
    df = pd.DataFrame({
        'contents': contents_list,
        'ids': ids_list,
        'scores': rerank_scores,
    })
    df[['contents', 'ids', 'scores']] = df.apply(sort_by_scores, axis=1, result_type='expand')
    results = select_top_k(df, ['contents', 'ids', 'scores'], top_k)

    del model
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return results['contents'].tolist(), results['ids'].tolist(), results['scores'].tolist()


def colbert_run_model(contents_list, model, tokenizer, device, batch_size: int):
    batch_contents_list = make_batch(contents_list, batch_size)
    results = []

    for batch_contents in batch_contents_list:
        flattened_batch_queries, flattened_batch_contents = map(list, zip(*batch_contents))

        # Tokenize both queries and contents together
        feature = tokenizer(flattened_batch_queries, flattened_batch_contents, padding=True, truncation=True,
                            return_tensors="pt").to(device)

        # Process the combined feature through the model in one go
        outputs = model(**feature)
        last_hidden_state = outputs.last_hidden_state

        # Split the embeddings back into query and content parts
        num_queries = len(flattened_batch_queries)
        query_embedding, content_embedding = last_hidden_state.split(
            [num_queries, last_hidden_state.size(1) - num_queries], dim=1)

        # Calculate cosine similarity between query and content embeddings
        sim_matrix = torch.nn.functional.cosine_similarity(
            query_embedding.unsqueeze(2), content_embedding.unsqueeze(1), dim=-1
        )

        # Take the maximum similarity for each query token (across all document tokens)
        max_sim_scores, _ = torch.max(sim_matrix, dim=2)
        results.extend(torch.mean(max_sim_scores, dim=1))

    return list(map(float, results))



import functools
import os
from pathlib import Path
from typing import Union, Tuple, List

import pandas as pd

from autorag.utils import result_to_dataframe, validate_qa_dataset, fetch_contents


# same with passage filter from now
def passage_filter_node(func):
    @functools.wraps(func)
    @result_to_dataframe(['retrieved_contents', 'retrieved_ids', 'retrieve_scores'])
    def wrapper(
            project_dir: Union[str, Path],
            previous_result: pd.DataFrame,
            *args, **kwargs) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
        validate_qa_dataset(previous_result)

        # find queries columns
        assert "query" in previous_result.columns, "previous_result must have query column."
        queries = previous_result["query"].tolist()

        # find contents_list columns
        assert "retrieved_contents" in previous_result.columns, "previous_result must have retrieved_contents column."
        contents = previous_result["retrieved_contents"].tolist()

        # find scores columns
        assert "retrieve_scores" in previous_result.columns, "previous_result must have retrieve_scores column."
        scores = previous_result["retrieve_scores"].tolist()

        # find ids columns
        assert "retrieved_ids" in previous_result.columns, "previous_result must have retrieved_ids column."
        ids = previous_result["retrieved_ids"].tolist()

        if func.__name__ == 'recency_filter':
            corpus_df = pd.read_parquet(os.path.join(project_dir, "data", "corpus.parquet"))
            metadatas = fetch_contents(corpus_df, ids, column_name='metadata')
            times = [[time['last_modified_datetime'] for time in time_list] for time_list in metadatas]
            filtered_contents, filtered_ids, filtered_scores \
                = func(contents_list=contents, scores_list=scores, ids_list=ids, time_list=times, *args, **kwargs)
        else:
            filtered_contents, filtered_ids, filtered_scores = func(queries=queries, contents_list=contents,
                                                                    scores_list=scores, ids_list=ids, *args, **kwargs)

        return filtered_contents, filtered_ids, filtered_scores

    return wrapper



from typing import List

import pandas as pd

queries = [
    "What is the capital of France?",
    "What is the meaning of life?",
]
retrieved_contents = [
    ["Paris is the capital of France.", "France is a country in Europe.", "France is a member of the EU."],
    ["The meaning of life is 42.", "The meaning of life is to be happy.", "The meaning of life is to be kind."],
]

df = pd.DataFrame({
    'query': queries,
    'retrieved_contents': retrieved_contents,
    'retrieved_ids': [['id-1', 'id-2', 'id-3'], ['id-4', 'id-5', 'id-6']],
    'retrieve_scores': [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]],
})


def check_result(result: List[str]):
    assert len(result) == len(queries)
    for r in result:
        assert isinstance(r, str)
        assert len(r) > 0
        assert bool(r) is True



import pandas as pd
import pytest

from autorag.nodes.retrieval import hybrid_rrf
from autorag.nodes.retrieval.hybrid_rrf import rrf_pure
from tests.autorag.nodes.retrieval.test_hybrid_base import (sample_ids, sample_scores, previous_result,
                                                            pseudo_project_dir)


def test_hybrid_rrf():
    result_id, result_scores = hybrid_rrf.__wrapped__(sample_ids, sample_scores, top_k=3, rrf_k=1)
    assert result_id == [
        ['id-3', 'id-1', 'id-2'],
        ['id-4', 'id-2', 'id-3']
    ]
    assert result_scores[0] == pytest.approx([1.0, (1 / 4) + (1 / 3), (1 / 3)])
    assert result_scores[1] == pytest.approx([1.0, (1 / 4) + (1 / 3), (1 / 3)])


def test_rrf_pure():
    sample_ids = (['id-1', 'id-2', 'id-3'],
                  ['id-1', 'id-4', 'id-3'])
    sample_scores = ([1, 3, 5], [4, 2, 6])
    result_id, result_scores = rrf_pure(sample_ids, sample_scores,
                                        rrf_k=1, top_k=3)
    assert result_scores == pytest.approx([1.0, (1 / 4) + (1 / 3), (1 / 3)])
    assert result_id == ['id-3', 'id-1', 'id-2']


def test_hybrid_rrf_node(pseudo_project_dir):
    modules = {
        'ids': ([['id-1', 'id-2', 'id-3'],
                 ['id-1', 'id-2', 'id-3'],
                 ['id-1', 'id-2', 'id-3']],
                [['id-7', 'id-8', 'id-9'],
                ['id-7', 'id-8', 'id-9'],
                ['id-7', 'id-8', 'id-9']]),
        'scores': ([
            [0.1, 0.2, 0.3],
            [0.1, 0.2, 0.3],
            [0.1, 0.2, 0.3]
        ],[
            [0.5, 0.6, 0.7],
            [0.5, 0.6, 0.7],
            [0.5, 0.6, 0.7]
        ]),
        'top_k': 3,
        'rrf_k': 1,
    }
    result_df = hybrid_rrf(project_dir=pseudo_project_dir, previous_result=previous_result, **modules)
    assert len(result_df) == 3
    assert isinstance(result_df, pd.DataFrame)
    assert set(result_df.columns) == {'retrieved_contents', 'retrieved_ids', 'retrieve_scores'}
    assert set(result_df['retrieved_ids'].tolist()[0]) == {'id-9', 'id-3', 'id-2'}
    assert result_df['retrieve_scores'].tolist()[0] == pytest.approx([0.5, 0.5, 1 / 3])
    assert set(result_df['retrieved_contents'].tolist()[0]) == {'doc-9', 'doc-3', 'doc-2'}


def test_hybrid_rrf_node_deploy(pseudo_project_dir):
    modules = {
        'target_modules': ('bm25', 'vectordb'),
        'target_module_params': [
            {'top_k': 3},
            {'embedding_model': 'openai', 'top_k': 3}
        ],
        'top_k': 3,
        'rrf_k': 1,
    }
    result_df = hybrid_rrf(project_dir=pseudo_project_dir, previous_result=previous_result, **modules)
    assert len(result_df) == 3
    assert isinstance(result_df, pd.DataFrame)
    assert set(result_df.columns) == {'retrieved_contents', 'retrieved_ids', 'retrieve_scores'}
    assert len(result_df['retrieved_ids'].tolist()[0]) == 3
    assert len(result_df['retrieve_scores'].tolist()[0]) == 3
    assert len(result_df['retrieved_contents'].tolist()[0]) == 3



from datetime import datetime

import numpy as np
import pandas as pd


def validate_qa_dataset(df: pd.DataFrame):
    columns = ['qid', 'query', 'retrieval_gt', 'generation_gt']
    assert set(columns).issubset(df.columns), f"df must have columns {columns}, but got {df.columns}"


def validate_corpus_dataset(df: pd.DataFrame):
    columns = ['doc_id', 'contents', 'metadata']
    assert set(columns).issubset(df.columns), f"df must have columns {columns}, but got {df.columns}"


def cast_qa_dataset(df: pd.DataFrame):
    def cast_retrieval_gt(gt):
        if isinstance(gt, str):
            return [[gt]]
        elif isinstance(gt, list):
            if isinstance(gt[0], str):
                return [gt]
            elif isinstance(gt[0], list):
                return gt
            elif isinstance(gt[0], np.ndarray):
                return cast_retrieval_gt(list(map(lambda x: x.tolist(), gt)))
            else:
                raise ValueError(f"retrieval_gt must be str or list, but got {type(gt[0])}")
        elif isinstance(gt, np.ndarray):
            return cast_retrieval_gt(gt.tolist())
        else:
            raise ValueError(f"retrieval_gt must be str or list, but got {type(gt)}")

    def cast_generation_gt(gt):
        if isinstance(gt, str):
            return [gt]
        elif isinstance(gt, list):
            return gt
        elif isinstance(gt, np.ndarray):
            return cast_generation_gt(gt.tolist())
        else:
            raise ValueError(f"generation_gt must be str or list, but got {type(gt)}")

    validate_qa_dataset(df)
    assert df['qid'].apply(lambda x: isinstance(x, str)).sum() == len(df), \
        "qid must be string type."
    assert df['query'].apply(lambda x: isinstance(x, str)).sum() == len(df), \
        "query must be string type."
    df['retrieval_gt'] = df['retrieval_gt'].apply(cast_retrieval_gt)
    df['generation_gt'] = df['generation_gt'].apply(cast_generation_gt)
    return df


def cast_corpus_dataset(df: pd.DataFrame):
    validate_corpus_dataset(df)

    def make_datetime_metadata(x):
        if x is None or x == {}:
            return {'last_modified_datetime': datetime.now()}
        elif x.get('last_modified_datetime') is None:
            return {**x, 'last_modified_datetime': datetime.now()}
        else:
            return x

    df['metadata'] = df['metadata'].apply(make_datetime_metadata)

    # check every metadata have a datetime key
    assert sum(df['metadata'].apply(lambda x: x.get('last_modified_datetime') is not None)) == len(df), \
        "Every metadata must have a datetime key."

    return df



import asyncio
from typing import List, Tuple

import numpy as np
import torch
from transformers import AutoModelForSequenceClassification, AutoTokenizer

from autorag.nodes.passagereranker.base import passage_reranker_node


@passage_reranker_node
def koreranker(queries: List[str], contents_list: List[List[str]],
               scores_list: List[List[float]], ids_list: List[List[str]],
               top_k: int) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents based on their relevance to a query using ko-reranker.
    ko-reranker is a reranker based on korean (https://huggingface.co/Dongjin-kr/ko-reranker).

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    model_path = "Dongjin-kr/ko-reranker"
    tokenizer = AutoTokenizer.from_pretrained(model_path)
    model = AutoModelForSequenceClassification.from_pretrained(model_path)
    model.eval()

    # Determine the device to run the model on (GPU if available, otherwise CPU)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)
    # Run async ko_rerank_pure function
    tasks = [koreranker_pure(query, contents, scores, ids, top_k, model, tokenizer, device)
             for query, contents, scores, ids in zip(queries, contents_list, scores_list, ids_list)]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(asyncio.gather(*tasks))
    content_result = list(map(lambda x: x[0], results))
    id_result = list(map(lambda x: x[1], results))
    score_result = list(map(lambda x: x[2], results))

    del model
    del tokenizer
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return content_result, id_result, score_result


async def koreranker_pure(query: str, contents: List[str],
                          scores: List[float], ids: List[str],
                          top_k: int, model, tokenizer, device) \
        -> Tuple[List[str], List[str], List[float]]:
    """
    Rerank a list of contents based on their relevance to a query using ko-reranker.

    :param query: The query to use for reranking
    :param contents: The list of contents to rerank
    :param scores: The list of scores retrieved from the initial ranking
    :param ids: The list of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param model: The ko-reranker model to use for reranking
    :param tokenizer: The tokenizer to use for the model
    :param device: The device to run the model on (GPU if available, otherwise CPU)
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    input_pairs = [[query, content] for content in contents]
    inputs = tokenizer(input_pairs, padding=True, truncation=True, return_tensors='pt', max_length=512)
    inputs = inputs.to(device)

    with torch.no_grad():
        scores = model(**inputs, return_dict=True).logits.view(-1, ).float()
        scores_np = scores.cpu().numpy()
        scores = exp_normalize(scores_np)

    # Convert scores type to float
    scores = scores.astype(float)

    # Create a list of tuples pairing each content with its relevance score
    content_ids_scores = list(zip(contents, ids, scores))

    # Sort the list of pairs based on the relevance score in descending order
    sorted_content_ids_scores = sorted(content_ids_scores, key=lambda x: x[2], reverse=True)

    # crop with top_k
    if len(contents) < top_k:
        top_k = len(contents)
    sorted_content_ids_scores = sorted_content_ids_scores[:top_k]

    content_result, id_result, score_result = zip(*sorted_content_ids_scores)

    return list(content_result), list(id_result), list(score_result)


def exp_normalize(x):
    b = x.max()
    y = np.exp(x - b)
    return y / y.sum()



from autorag.nodes.passagereranker import pass_reranker
from tests.autorag.nodes.passagereranker.test_passage_reranker_base \
    import project_dir, previous_result, ids_example


def test_pass_reranker():
    top_k = 1
    result_df = pass_reranker(project_dir=project_dir, previous_result=previous_result, top_k=top_k)
    assert 'retrieved_contents' in result_df.columns
    assert 'retrieved_ids' in result_df.columns
    assert 'retrieve_scores' in result_df.columns

    contents = result_df["retrieved_contents"].tolist()
    assert contents == [['NomaDamas is Great Team'],
                        ['i am hungry']]
    ids = result_df["retrieved_ids"].tolist()
    assert ids == [[ids_example[0][0]], [ids_example[1][0]]]
    scores = result_df["retrieve_scores"].tolist()
    assert scores == [[0.1], [0.1]]



import math
from typing import Tuple, List

import pandas as pd

from autorag.evaluate import evaluate_retrieval

retrieval_gt = [[[f'test{i}-{j}'] for i in range(2)] for j in range(4)]


@evaluate_retrieval(retrieval_gt=retrieval_gt, metrics=['retrieval_recall', 'retrieval_precision', 'retrieval_f1'])
def pseudo_retrieval() -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    contents = [
        ['a', 'b', 'c', 'd'],
        ['better', 'bone', 'caleb', 'done'],
        ['ate', 'better', 'cortex', 'dad'],
        ['aim', 'bond', 'cane', 'dance'],
    ]
    scores = [
        [0.1, 0.2, 0.3, 0.4],
        [0.4, 0.3, 0.2, 0.1],
        [0.5, 0.4, 0.3, 0.2],
        [0.6, 0.5, 0.4, 0.3],
    ]
    ids = [
        [retrieval_gt[0][0][0], retrieval_gt[0][1][0], 'pred-0', 'pred-1'],
        ['pred-2', retrieval_gt[1][0][0], retrieval_gt[1][1][0], 'pred-3'],
        [f'pred-{i}' for i in range(4, 8)],
        [retrieval_gt[3][0][0], 'pred-8', 'pred-9', 'pred-10'],
    ]
    return contents, ids, scores


def test_evaluate_retrieval():
    result_df = pseudo_retrieval()
    assert isinstance(result_df, pd.DataFrame)
    assert len(result_df) == 4
    assert len(result_df.columns) == 6
    assert list(result_df.columns) == ['retrieved_contents', 'retrieved_ids', 'retrieve_scores', 'retrieval_recall',
                                       'retrieval_precision', 'retrieval_f1']
    recall = result_df['retrieval_recall'].tolist()
    recall_solution = [1, 1, 0, 0.5]
    for pred, sol in zip(recall, recall_solution):
        assert math.isclose(pred, sol, rel_tol=1e-4)

    precision = result_df['retrieval_precision'].tolist()
    precision_solution = [0.5, 0.5, 0, 0.25]
    for pred, sol in zip(precision, precision_solution):
        assert math.isclose(pred, sol, rel_tol=1e-4)

    f1 = result_df['retrieval_f1'].tolist()
    f1_solution = [1 / 1.5, 1 / 1.5, 0, 0.25 / 0.75]
    for pred, sol in zip(f1, f1_solution):
        assert math.isclose(pred, sol, rel_tol=1e-4)



import functools
import logging
import os
import pickle
from pathlib import Path
from typing import List, Union, Tuple, Dict

import chromadb
import pandas as pd
import torch

from autorag import embedding_models
from autorag.support import get_support_modules
from autorag.utils import fetch_contents, result_to_dataframe, validate_qa_dataset

logger = logging.getLogger("AutoRAG")


def retrieval_node(func):
    """
    Load resources for running retrieval_node.
    For example, it loads bm25 corpus for bm25 retrieval.

    :param func: Retrieval function that returns a list of ids and a list of scores
    :return: A pandas Dataframe that contains retrieved contents, retrieved ids, and retrieve scores.
        The column name will be "retrieved_contents", "retrieved_ids", and "retrieve_scores".
    """

    @functools.wraps(func)
    @result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
    def wrapper(
            project_dir: Union[str, Path],
            previous_result: pd.DataFrame,
            **kwargs) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
        validate_qa_dataset(previous_result)
        resources_dir = os.path.join(project_dir, "resources")
        data_dir = os.path.join(project_dir, "data")

        if func.__name__ == "bm25":
            # check if bm25_path and file exists
            bm25_path = os.path.join(resources_dir, 'bm25.pkl')
            assert bm25_path is not None, "bm25_path must be specified for using bm25 retrieval."
            assert os.path.exists(bm25_path), f"bm25_path {bm25_path} does not exist. Please ingest first."
        elif func.__name__ == "vectordb":
            # check if chroma_path and file exist
            chroma_path = os.path.join(resources_dir, 'chroma')
            embedding_model_str = kwargs.pop("embedding_model")
            assert chroma_path is not None, "chroma_path must be specified for using vectordb retrieval."
            assert os.path.exists(chroma_path), f"chroma_path {chroma_path} does not exist. Please ingest first."

        # find queries columns & type cast queries
        assert "query" in previous_result.columns, "previous_result must have query column."
        if "queries" not in previous_result.columns:
            previous_result["queries"] = previous_result["query"]
        previous_result["queries"] = previous_result["queries"].apply(cast_queries)
        queries = previous_result["queries"].tolist()

        # run retrieval function
        if func.__name__ == "bm25":
            bm25_corpus = load_bm25_corpus(bm25_path)
            ids, scores = func(queries=queries, bm25_corpus=bm25_corpus, **kwargs)
        elif func.__name__ == "vectordb":
            chroma_collection = load_chroma_collection(db_path=chroma_path, collection_name=embedding_model_str)
            if embedding_model_str in embedding_models:
                embedding_model = embedding_models[embedding_model_str]
            else:
                logger.error(f"embedding_model_str {embedding_model_str} does not exist.")
                raise KeyError(f"embedding_model_str {embedding_model_str} does not exist.")
            ids, scores = func(queries=queries, collection=chroma_collection,
                               embedding_model=embedding_model, **kwargs)
            del embedding_model
            if torch.cuda.is_available():
                torch.cuda.empty_cache()
        elif func.__name__ in ["hybrid_rrf", "hybrid_cc", "hybrid_rsf", "hybrid_dbsf"]:
            if 'ids' in kwargs and 'scores' in kwargs:
                ids, scores = func(**kwargs)
            else:
                if not ('target_modules' in kwargs and 'target_module_params' in kwargs):
                    raise ValueError(
                        f"If there are no ids and scores specified, target_modules and target_module_params must be specified for using {func.__name__}.")
                target_modules = kwargs.pop('target_modules')
                target_module_params = kwargs.pop('target_module_params')
                result_dfs = list(map(lambda x: get_support_modules(x[0])(**x[1], project_dir=project_dir,
                                                                          previous_result=previous_result),
                                      zip(target_modules, target_module_params)))
                ids = tuple(map(lambda df: df['retrieved_ids'].apply(list).tolist(), result_dfs))
                scores = tuple(map(lambda df: df['retrieve_scores'].apply(list).tolist(), result_dfs))
                ids, scores = func(ids=ids, scores=scores, **kwargs)
        else:
            raise ValueError(f"invalid func name for using retrieval_io decorator.")

        # fetch data from corpus_data
        corpus_data = pd.read_parquet(os.path.join(data_dir, "corpus.parquet"))
        contents = fetch_contents(corpus_data, ids)

        return contents, ids, scores

    return wrapper


def load_bm25_corpus(bm25_path: str) -> Dict:
    if bm25_path is None:
        return {}
    with open(bm25_path, "rb") as f:
        bm25_corpus = pickle.load(f)
    return bm25_corpus


def load_chroma_collection(db_path: str, collection_name: str) -> chromadb.Collection:
    db = chromadb.PersistentClient(path=db_path)
    collection = db.get_collection(name=collection_name)
    return collection


def cast_queries(queries: Union[str, List[str]]) -> List[str]:
    if isinstance(queries, str):
        return [queries]
    elif isinstance(queries, List):
        return queries
    else:
        raise ValueError(f"queries must be str or list, but got {type(queries)}")


def evenly_distribute_passages(ids: List[List[str]], scores: List[List[float]], top_k: int) -> Tuple[
    List[str], List[float]]:
    assert len(ids) == len(scores), "ids and scores must have same length."
    query_cnt = len(ids)
    avg_len = top_k // query_cnt
    remainder = top_k % query_cnt

    new_ids = []
    new_scores = []
    for i in range(query_cnt):
        if i < remainder:
            new_ids.extend(ids[i][:avg_len + 1])
            new_scores.extend(scores[i][:avg_len + 1])
        else:
            new_ids.extend(ids[i][:avg_len])
            new_scores.extend(scores[i][:avg_len])

    return new_ids, new_scores


def run_retrieval_modules(project_dir: str, previous_result: pd.DataFrame,
                          module_name: str, module_params: Dict) -> pd.DataFrame:
    return



from typing import List, Tuple

import pandas as pd

from autorag.nodes.retrieval.base import retrieval_node
from autorag.nodes.retrieval.hybrid_rsf import score_fusion


@retrieval_node
def hybrid_dbsf(ids: Tuple, scores: Tuple, top_k: int,
                weights: Tuple = (0.5, 0.5)) -> Tuple[List[str], List[float]]:
    """
    Hybrid DBSF function.
    DBSF (Distribution Based Score Fusion) is a method to combine multiple retrieval results based on their distribution.
    It is uniquer than other retrieval modules, because it does not really execute retrieval,
    but just fuse the results of other retrieval functions.
    So you have to run more than two retrieval modules before running this function.
    And collect ids and scores result from each retrieval module.
    Make it as tuple and input it to this function.

    :param ids: The tuple of ids that you want to fuse.
        The length of this must be the same as the length of scores.
    :param scores: The retrieve scores that you want to fuse.
        The length of this must be the same as the length of ids.
    :param top_k: The number of passages to be retrieved.
    :param weights: Weight for each retrieval result.
        Default is (0.5, 0.5).
        You must set its length as the same as the length of ids and scores.
        Plus, the sum of the weights must be 1.
    :return: The tuple of ids and fused scores that fused by DBSF.
    """
    assert len(ids) == len(scores), "The length of ids and scores must be the same."
    assert len(ids) > 1, "You must input more than one retrieval results."
    assert top_k > 0, "top_k must be greater than 0."
    assert sum(weights) == 1, "The sum of weights must be 1."

    # Initialize DataFrame for ids and scores
    id_df = pd.DataFrame({f'id_{i}': id_list for i, id_list in enumerate(ids)})
    score_df = pd.DataFrame({f'score_{i}': score_list for i, score_list in enumerate(scores)})
    df = pd.concat([id_df, score_df], axis=1)

    # Apply relative score fusion
    def dbsf_pure_apply(row):
        ids_tuple = tuple(row[[f'id_{i}' for i in range(len(ids))]].values)
        scores_tuple = tuple(row[[f'score_{i}' for i in range(len(scores))]].values)
        return pd.Series(dbsf_pure(ids_tuple, scores_tuple, top_k, weights))

    df[['dbsf_id', 'dbsf_score']] = df.apply(dbsf_pure_apply, axis=1)
    return df['dbsf_id'].tolist(), df['dbsf_score'].tolist()


def dbsf_pure(ids: Tuple, scores: Tuple, top_k: int,
              weights: Tuple = (0.5, 0.5)) -> Tuple[List[str], List[float]]:
    min_max_scores = dbsf_min_max_scores(scores)
    top_ids, top_scores = score_fusion(ids, scores, top_k, min_max_scores, weights)
    return top_ids, top_scores


def dbsf_min_max_scores(scores: Tuple):
    min_max_scores = {}
    for i, score_list in enumerate(scores):
        if not score_list:
            min_max_scores[i] = (0.0, 0.0)
            continue
        mean_score = sum(score_list) / len(score_list)
        std_dev = (sum((x - mean_score) ** 2 for x in score_list) / len(score_list)) ** 0.5
        min_score = mean_score - 3 * std_dev
        max_score = mean_score + 3 * std_dev
        min_max_scores[i] = (min_score, max_score)
    return min_max_scores



import importlib
from typing import Callable, Dict


def dynamically_find_function(key: str, target_dict: Dict) -> Callable:
    if key in target_dict:
        module_path, func_name = target_dict[key]
        module = importlib.import_module(module_path)
        func = getattr(module, func_name)
        return func
    else:
        raise KeyError(f"Input module or node {key} is not supported.")


def get_support_modules(module_name: str) -> Callable:
    support_modules = {
        # query_expansion
        'query_decompose': ('autorag.nodes.queryexpansion', 'query_decompose'),
        'hyde': ('autorag.nodes.queryexpansion', 'hyde'),
        'pass_query_expansion': ('autorag.nodes.queryexpansion', 'pass_query_expansion'),
        # retrieval
        'bm25': ('autorag.nodes.retrieval', 'bm25'),
        'vectordb': ('autorag.nodes.retrieval', 'vectordb'),
        'hybrid_rrf': ('autorag.nodes.retrieval', 'hybrid_rrf'),
        'hybrid_cc': ('autorag.nodes.retrieval', 'hybrid_cc'),
        'hybrid_rsf': ('autorag.nodes.retrieval', 'hybrid_rsf'),
        'hybrid_dbsf': ('autorag.nodes.retrieval', 'hybrid_dbsf'),
        # passage_reranker
        'monot5': ('autorag.nodes.passagereranker', 'monot5'),
        'tart': ('autorag.nodes.passagereranker', 'tart'),
        'upr': ('autorag.nodes.passagereranker', 'upr'),
        'koreranker': ('autorag.nodes.passagereranker', 'koreranker'),
        'pass_reranker': ('autorag.nodes.passagereranker', 'pass_reranker'),
        'cohere_reranker': ('autorag.nodes.passagereranker', 'cohere_reranker'),
        'rankgpt': ('autorag.nodes.passagereranker', 'rankgpt'),
        'jina_reranker': ('autorag.nodes.passagereranker', 'jina_reranker'),
        'colbert_reranker': ('autorag.nodes.passagereranker', 'colbert_reranker'),
        'sentence_transformer_reranker': ('autorag.nodes.passagereranker', 'sentence_transformer_reranker'),
        'flag_embedding_reranker': ('autorag.nodes.passagereranker', 'flag_embedding_reranker'),
        'flag_embedding_llm_reranker': ('autorag.nodes.passagereranker', 'flag_embedding_llm_reranker'),
        'time_reranker': ('autorag.nodes.passagereranker', 'time_reranker'),
        # passage_filter
        'pass_passage_filter': ('autorag.nodes.passagefilter', 'pass_passage_filter'),
        'similarity_threshold_cutoff': ('autorag.nodes.passagefilter', 'similarity_threshold_cutoff'),
        'similarity_percentile_cutoff': ('autorag.nodes.passagefilter', 'similarity_percentile_cutoff'),
        'recency_filter': ('autorag.nodes.passagefilter', 'recency_filter'),
        # passage_compressor
        'tree_summarize': ('autorag.nodes.passagecompressor', 'tree_summarize'),
        'pass_compressor': ('autorag.nodes.passagecompressor', 'pass_compressor'),
        'refine': ('autorag.nodes.passagecompressor', 'refine'),
        # prompt_maker
        'fstring': ('autorag.nodes.promptmaker', 'fstring'),
        'long_context_reorder': ('autorag.nodes.promptmaker', 'long_context_reorder'),
        # generator
        'llama_index_llm': ('autorag.nodes.generator', 'llama_index_llm'),
        'vllm': ('autorag.nodes.generator', 'vllm'),
    }
    return dynamically_find_function(module_name, support_modules)


def get_support_nodes(node_name: str) -> Callable:
    support_nodes = {
        'query_expansion': ('autorag.nodes.queryexpansion.run', 'run_query_expansion_node'),
        'retrieval': ('autorag.nodes.retrieval.run', 'run_retrieval_node'),
        'generator': ('autorag.nodes.generator.run', 'run_generator_node'),
        'prompt_maker': ('autorag.nodes.promptmaker.run', 'run_prompt_maker_node'),
        'passage_filter': ('autorag.nodes.passagefilter.run', 'run_passage_filter_node'),
        'passage_compressor': ('autorag.nodes.passagecompressor.run', 'run_passage_compressor_node'),
        'passage_reranker': ('autorag.nodes.passagereranker.run', 'run_passage_reranker_node'),
    }
    return dynamically_find_function(node_name, support_nodes)
